@@ -1,12 +1,7 @@
 //! Services for handling JSON-RPC requests.
 //!
-//! Requests handled by the server are first routed via the
-//! primary `ServiceHandler` and any response is sent back to the
-//! client making the request; these requests may mutate the server state.
-//!
-//! Afterwards the same request is sent to the `NotifyHandler` which
-//! must **never mutate the server state** but may send notifications
-//! to connected clients based on the updated server state.
+//! Some methods send notifications to connected clients, see the
+//! methods documentation for more details.
 //!
 //! Notifications sent to connected clients are sent as a tuple
 //! of `String` event name followed by an arbitrary JSON `Value`
@@ -262,7 +257,8 @@ pub struct ServiceHandler;
 
 #[async_trait]
 impl Service for ServiceHandler {
-    type Data = (usize, Arc<RwLock<State>>);
+    type Data = (usize, Arc<RwLock<State>>, Arc<Mutex<Option<Notification>>>);
+
     async fn handle(
         &self,
         req: &Request,
@@ -270,7 +266,7 @@ impl Service for ServiceHandler {
     ) -> Result<Option<Response>> {
         let response = match req.method() {
             GROUP_CREATE => {
-                let (conn_id, state) = ctx;
+                let (conn_id, state, _) = ctx;
                 let params: GroupCreateParams = req.deserialize()?;
                 let (label, parameters) = params;
 
@@ -302,7 +298,7 @@ impl Service for ServiceHandler {
                 Some((req, res).into())
             }
             GROUP_JOIN => {
-                let (conn_id, state) = ctx;
+                let (conn_id, state, _) = ctx;
                 let group_id: Uuid = req.deserialize()?;
                 let mut writer = state.write().await;
                 if let Some(group) = writer.groups.get_mut(&group_id) {
@@ -329,7 +325,7 @@ impl Service for ServiceHandler {
                 }
             }
             SESSION_CREATE => {
-                let (conn_id, state) = ctx;
+                let (conn_id, state, notification) = ctx;
                 let params: SessionCreateParams = req.deserialize()?;
                 let (group_id, kind) = params;
                 let mut writer = state.write().await;
@@ -338,11 +334,28 @@ impl Service for ServiceHandler {
                 let session = Session::from(kind.clone());
                 let key = session.uuid.clone();
                 group.sessions.insert(key, session.clone());
+
+                if let SessionKind::Keygen = kind {
+                    let value =
+                        serde_json::to_value((SESSION_CREATE_EVENT, &session))
+                            .unwrap();
+                    let response: Response = value.into();
+
+                    // Notify everyone else in the group a session was created
+                    let ctx = Notification::Group {
+                        group_id,
+                        filter: Some(vec![*conn_id]),
+                        response,
+                    };
+                    let mut writer = notification.lock().await;
+                    *writer = Some(ctx);
+                }
+
                 let res = serde_json::to_value(&session).unwrap();
                 Some((req, res).into())
             }
             SESSION_JOIN => {
-                let (conn_id, state) = ctx;
+                let (conn_id, state, _) = ctx;
                 let params: SessionJoinParams = req.deserialize()?;
                 let (group_id, session_id, _kind) = params;
 
@@ -359,15 +372,41 @@ impl Service for ServiceHandler {
                 }
             }
             SESSION_SIGNUP => {
-                let (conn_id, state) = ctx;
+                let (conn_id, state, notification) = ctx;
                 let params: SessionSignupParams = req.deserialize()?;
-                let (group_id, session_id, _kind) = params;
+                let (group_id, session_id, kind) = params;
 
                 let mut writer = state.write().await;
                 let group =
                     get_group_mut(&conn_id, &group_id, &mut writer.groups)?;
                 if let Some(session) = group.sessions.get_mut(&session_id) {
                     let party_number = session.signup(*conn_id);
+
+                    tracing::info!(party_number, "session signup {}", conn_id);
+
+                    // Enough parties are signed up to the session
+                    if threshold(
+                        &kind,
+                        &group.params,
+                        session.party_signups.len(),
+                    ) {
+                        let value = serde_json::to_value((
+                            SESSION_SIGNUP_EVENT,
+                            &session_id,
+                        ))
+                        .unwrap();
+                        let response: Response = value.into();
+                        let ctx = Notification::Session {
+                            group_id,
+                            session_id,
+                            filter: None,
+                            response,
+                        };
+
+                        let mut writer = notification.lock().await;
+                        *writer = Some(ctx);
+                    }
+
                     let res = serde_json::to_value(&party_number).unwrap();
                     Some((req, res).into())
                 } else {
@@ -379,7 +418,7 @@ impl Service for ServiceHandler {
             // Load an existing party signup into the session
             // this is used to support loading existing key shares.
             SESSION_LOAD => {
-                let (conn_id, state) = ctx;
+                let (conn_id, state, notification) = ctx;
                 let params: SessionLoadParams = req.deserialize()?;
                 let (group_id, session_id, kind, party_number) = params;
 
@@ -388,6 +427,28 @@ impl Service for ServiceHandler {
                     let group =
                         get_group_mut(&conn_id, &group_id, &mut writer.groups)?;
                     if let Some(session) = group.sessions.get_mut(&session_id) {
+                        // Enough parties are loaded into the session
+                        if threshold(
+                            &kind,
+                            &group.params,
+                            session.party_signups.len(),
+                        ) {
+                            let value = serde_json::to_value((
+                                SESSION_LOAD_EVENT,
+                                &session_id,
+                            ))
+                            .unwrap();
+                            let response: Response = value.into();
+                            let ctx = Notification::Session {
+                                group_id,
+                                session_id,
+                                filter: None,
+                                response,
+                            };
+                            let mut writer = notification.lock().await;
+                            *writer = Some(ctx);
+                        }
+
                         let res = serde_json::to_value(&party_number).unwrap();
                         match session.load(
                             &group.params,
@@ -412,7 +473,7 @@ impl Service for ServiceHandler {
             }
             // Mark the session as finished for a party.
             SESSION_FINISH => {
-                let (conn_id, state) = ctx;
+                let (conn_id, state, notification) = ctx;
                 let params: SessionFinishParams = req.deserialize()?;
                 let (group_id, session_id, party_number) = params;
 
@@ -435,6 +496,40 @@ impl Service for ServiceHandler {
                         }
 
                         session.finished.insert(party_number);
+
+                        let mut signups = session
+                            .party_signups
+                            .iter()
+                            .map(|(n, _)| n.clone())
+                            .collect::<Vec<u16>>();
+                        let mut completed = session
+                            .finished
+                            .iter()
+                            .cloned()
+                            .collect::<Vec<u16>>();
+
+                        signups.sort();
+                        completed.sort();
+
+                        if signups == completed {
+                            let value = serde_json::to_value((
+                                SESSION_CLOSED_EVENT,
+                                completed,
+                            ))
+                            .unwrap();
+                            let response: Response = value.into();
+
+                            let ctx = Notification::Session {
+                                group_id,
+                                session_id,
+                                filter: None,
+                                response,
+                            };
+
+                            let mut writer = notification.lock().await;
+                            *writer = Some(ctx);
+                        }
+
                         Some(req.into())
                     } else {
                         return Err(Error::from(Box::from(
@@ -447,119 +542,13 @@ impl Service for ServiceHandler {
                     )));
                 }
             }
-            SESSION_MESSAGE | NOTIFY_PROPOSAL | NOTIFY_SIGNED => {
-                // Must ACK so we indicate the service method exists
-                // the actual logic is handled by the notification service
-                Some(req.into())
-            }
-            _ => None,
-        };
-        Ok(response)
-    }
-}
-
-/// Service for broadcasting notifications to connected clients.
-pub struct NotifyHandler;
-
-#[async_trait]
-impl Service for NotifyHandler {
-    type Data = (usize, Arc<RwLock<State>>, Arc<Mutex<Notification>>);
-    async fn handle(
-        &self,
-        req: &Request,
-        ctx: &Self::Data,
-    ) -> Result<Option<Response>> {
-        let response = match req.method() {
-            SESSION_CREATE => {
-                let (conn_id, state, notification) = ctx;
-                let params: SessionCreateParams = req.deserialize()?;
-                let (group_id, kind) = params;
-
-                if let SessionKind::Keygen = kind {
-                    let reader = state.read().await;
-                    let group = get_group(&conn_id, &group_id, &reader.groups)?;
-
-                    let last_session =
-                        group.sessions.values().last().unwrap().clone();
-                    let res = serde_json::to_value((
-                        SESSION_CREATE_EVENT,
-                        &last_session,
-                    ))
-                    .unwrap();
-
-                    // Notify everyone else in the group a session was created
-                    {
-                        let ctx = Notification::Group {
-                            group_id,
-                            filter: Some(vec![*conn_id]),
-                        };
-                        let mut writer = notification.lock().await;
-                        *writer = ctx;
-                    }
-
-                    Some(res.into())
-                } else {
-                    return Err(Error::from(Box::from(
-                        ServiceError::KeygenSessionExpected,
-                    )));
-                }
-            }
-            SESSION_SIGNUP => {
-                let (conn_id, state, notification) = ctx;
-                let params: SessionSignupParams = req.deserialize()?;
-                let (group_id, session_id, kind) = params;
-
-                let reader = state.read().await;
-
-                let (group, session) = get_group_session(
-                    &conn_id,
-                    &group_id,
-                    &session_id,
-                    &reader.groups,
-                )?;
-                handle_threshold_notify(
-                    session.party_signups.len(),
-                    group_id,
-                    session_id,
-                    group,
-                    session,
-                    kind,
-                    notification,
-                    SESSION_SIGNUP_EVENT,
-                )
-                .await
-            }
-            SESSION_LOAD => {
-                let (conn_id, state, notification) = ctx;
-                let params: SessionLoadParams = req.deserialize()?;
-                let (group_id, session_id, kind, _party_number) = params;
-
-                let reader = state.read().await;
-
-                let (group, session) = get_group_session(
-                    &conn_id,
-                    &group_id,
-                    &session_id,
-                    &reader.groups,
-                )?;
-                handle_threshold_notify(
-                    session.party_signups.len(),
-                    group_id,
-                    session_id,
-                    group,
-                    session,
-                    kind,
-                    notification,
-                    SESSION_LOAD_EVENT,
-                )
-                .await
-            }
             SESSION_MESSAGE => {
                 let (conn_id, state, notification) = ctx;
                 let params: SessionMessageParams = req.deserialize()?;
                 let (group_id, session_id, _kind, msg) = params;
 
                 let reader = state.read().await;
+
                 // Check we have valid group / session
                 let (_group, session) = get_group_session(
                     &conn_id,
@@ -573,97 +562,44 @@ impl Service for NotifyHandler {
                     if let Some(s) =
                         session.party_signups.iter().find(|s| s.0 == *receiver)
                     {
-                        let result =
+                        let value =
                             serde_json::to_value((SESSION_MESSAGE_EVENT, msg))
                                 .unwrap();
 
-                        let response: Response = result.into();
+                        let response: Response = value.into();
                         let message = (s.1, response);
 
-                        {
-                            let ctx = Notification::Relay {
-                                messages: vec![message],
-                            };
+                        let ctx = Notification::Relay {
+                            messages: vec![message],
+                        };
 
-                            let mut writer = notification.lock().await;
-                            *writer = ctx;
-                        }
-
-                        // Must return a response so the server processes
-                        // our notifications even though our actual responses
-                        // are in the messages assigned to the notification context
-                        Some((serde_json::Value::Null).into())
+                        let mut writer = notification.lock().await;
+                        *writer = Some(ctx);
                     } else {
                         return Err(Error::from(Box::from(
                             ServiceError::BadPeerReceiver(*receiver),
                         )));
                     }
-
                 // Handle broadcast round
                 } else {
-                    {
-                        let ctx = Notification::Session {
-                            group_id,
-                            session_id,
-                            filter: Some(vec![*conn_id]),
-                        };
-
-                        let mut writer = notification.lock().await;
-                        *writer = ctx;
-                    }
-
-                    let result =
+                    let value =
                         serde_json::to_value((SESSION_MESSAGE_EVENT, msg))
                             .unwrap();
+                    let response: Response = value.clone().into();
 
-                    Some(result.into())
+                    let ctx = Notification::Session {
+                        group_id,
+                        session_id,
+                        filter: Some(vec![*conn_id]),
+                        response,
+                    };
+
+                    let mut writer = notification.lock().await;
+                    *writer = Some(ctx);
                 }
-            }
-            SESSION_FINISH => {
-                let (conn_id, state, notification) = ctx;
-                let params: SessionFinishParams = req.deserialize()?;
-                let (group_id, session_id, _party_number) = params;
 
-                let reader = state.read().await;
-
-                let (_group, session) = get_group_session(
-                    &conn_id,
-                    &group_id,
-                    &session_id,
-                    &reader.groups,
-                )?;
-
-                let mut signups = session
-                    .party_signups
-                    .iter()
-                    .map(|(n, _)| n.clone())
-                    .collect::<Vec<u16>>();
-                let mut completed =
-                    session.finished.iter().cloned().collect::<Vec<u16>>();
-
-                signups.sort();
-                completed.sort();
-
-                if signups == completed {
-                    let result =
-                        serde_json::to_value((SESSION_CLOSED_EVENT, completed))
-                            .unwrap();
-
-                    {
-                        let ctx = Notification::Session {
-                            group_id,
-                            session_id,
-                            filter: None,
-                        };
-
-                        let mut writer = notification.lock().await;
-                        *writer = ctx;
-                    }
-
-                    Some(result.into())
-                } else {
-                    None
-                }
+                // Must ACK so we indicate the service method exists
+                Some(req.into())
             }
             NOTIFY_PROPOSAL => {
                 let (conn_id, _state, notification) = ctx;
@@ -676,22 +612,24 @@ impl Service for NotifyHandler {
                     message,
                 };
 
-                let res =
+                let value =
                     serde_json::to_value((NOTIFY_PROPOSAL_EVENT, &proposal))
                         .unwrap();
+                let response: Response = value.into();
 
-                {
-                    let ctx = Notification::Group {
-                        group_id,
-                        filter: Some(vec![*conn_id]),
-                    };
+                let ctx = Notification::Group {
+                    group_id,
+                    filter: Some(vec![*conn_id]),
+                    response,
+                };
 
-                    let mut writer = notification.lock().await;
-                    *writer = ctx;
-                }
+                let mut writer = notification.lock().await;
+                *writer = Some(ctx);
 
-                Some(res.into())
+                // Must ACK so we indicate the service method exists
+                Some(req.into())
             }
+
             NOTIFY_SIGNED => {
                 let (conn_id, state, notification) = ctx;
                 let params: NotifySignedParams = req.deserialize()?;
@@ -712,23 +650,24 @@ impl Service for NotifyHandler {
                     .map(|(_, c)| c.clone())
                     .collect::<Vec<usize>>();
 
-                let result =
+                let value =
                     serde_json::to_value((NOTIFY_SIGNED_EVENT, value)).unwrap();
+                let response: Response = value.into();
 
-                {
-                    let ctx = Notification::Group {
-                        group_id,
-                        filter: Some(participants),
-                    };
+                let ctx = Notification::Group {
+                    group_id,
+                    filter: Some(participants),
+                    response,
+                };
 
-                    let mut writer = notification.lock().await;
-                    *writer = ctx;
-                }
-                Some(result.into())
+                let mut writer = notification.lock().await;
+                *writer = Some(ctx);
+
+                // Must ACK so we indicate the service method exists
+                Some(req.into())
             }
             _ => None,
         };
-
         Ok(response)
     }
 }
@@ -793,48 +732,17 @@ fn get_group_session<'a>(
     }
 }
 
-async fn handle_threshold_notify(
+/// Helper to determine if we met a session party threshold.
+fn threshold(
+    kind: &SessionKind,
+    params: &Parameters,
     num_entries: usize,
-    group_id: String,
-    session_id: String,
-    group: &Group,
-    _session: &Session,
-    kind: SessionKind,
-    notification: &Mutex<Notification>,
-    event: &str,
-) -> Option<Response> {
-    let parties = group.params.parties as usize;
-    let threshold = group.params.threshold as usize;
+) -> bool {
+    let parties = params.parties as usize;
+    let threshold = params.threshold as usize;
     let required_num_entries = match kind {
         SessionKind::Keygen => parties,
         SessionKind::Sign => threshold + 1,
     };
-
-    // Enough parties are signed up to the session
-    if num_entries == required_num_entries {
-        println!("Sending threshold notify event {} for {:#?}", event, kind);
-
-        let res = serde_json::to_value((event, &session_id)).unwrap();
-
-        // Notify everyone in the session that enough
-        // parties have signed up to the session
-        {
-            let ctx = Notification::Session {
-                group_id,
-                session_id,
-                filter: None,
-            };
-
-            let mut writer = notification.lock().await;
-            *writer = ctx;
-        }
-
-        Some(res.into())
-    } else {
-        {
-            let mut writer = notification.lock().await;
-            *writer = Default::default();
-        }
-        None
-    }
+    num_entries == required_num_entries
 }
